/* ====================================================================
 * Copyright (c) 1996-2000 The Apache Software Foundation.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Software Foundation
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Software Foundation" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Software Foundation.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Software Foundation
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE Apache Software Foundation ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE Apache Software Foundation OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Software Foundation and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

#include "ap_config.h"
#include "httpd.h"
#include "http_main.h"
#include "http_log.h"

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>

#ifdef HAVE_BSTRING_H
#include <bstring.h>		/* for IRIX, FD_SET calls bzero() */
#endif

#ifndef DEFAULT_BUFSIZE
#define DEFAULT_BUFSIZE (4096)
#endif

/* the maximum size of any chunk */
#ifndef MAX_CHUNK_SIZE
#define MAX_CHUNK_SIZE (0x8000)
#endif

/* This must be enough to represent MAX_CHUNK_SIZE in hex,
 * plus two extra characters.
 */
#ifndef CHUNK_HEADER_SIZE
#define CHUNK_HEADER_SIZE (6)
#endif

/* bwrite()s of greater than this size can result in a large_write() call,
 * which can result in a writev().  It's a little more work to set up the
 * writev() rather than copy bytes into the buffer, so we don't do it for small
 * writes.  This is especially important when chunking (which is a very likely
 * source of small writes if it's a module using ap_bputc/ap_bputs)...because we
 * have the expense of actually building two chunks for each writev().
 */
#ifndef LARGE_WRITE_THRESHOLD
#define LARGE_WRITE_THRESHOLD 31
#endif

/*
 * Buffered I/O routines.
 * These are a replacement for the stdio routines.
 * Advantages:
 *  Known semantics for handling of file-descriptors (on close etc.)
 *  No problems reading and writing simultanously to the same descriptor
 *  No limits on the number of open file handles.
 *  Only uses memory resources; no need to ensure the close routine
 *  is called.
 *  Extra code could be inserted between the buffered and un-buffered routines.
 *  Timeouts could be handled by using select or poll before read or write.
 *  Extra error handling could be introduced; e.g.
 *   keep an address to which we should longjump(), or
 *   keep a stack of routines to call on error.
 */

/* Notes:
 *  On reading EOF, EOF will set in the flags and no further Input will
 * be done.
 *
 * On an error except for EAGAIN, ERROR will be set in the flags and no
 * futher I/O will be done
 */


static void doerror(BUFF *fb, int direction)
{
    fb->flags |= (direction == B_RD ? B_RDERR : B_WRERR);
    if (fb->error != NULL)
	(*fb->error) (fb, direction, fb->error_data, fb->saved_errno);
}

/* Buffering routines */
/*
 * Create a new buffered stream
 */
API_EXPORT(BUFF *) ap_bcreate(ap_context_t *p, int flags)
{
    BUFF *fb;

    fb = ap_palloc(p, sizeof(BUFF));
    fb->pool = p;
    fb->bufsiz = DEFAULT_BUFSIZE;
    fb->flags = flags & B_RDWR;

    if (flags & B_RD)
	fb->inbase = ap_palloc(p, fb->bufsiz);
    else
	fb->inbase = NULL;

    /* overallocate so that we can put a chunk trailer of CRLF into this
     * buffer... and possibly the beginning of a new chunk
     */
    if (flags & B_WR)
	fb->outbase = ap_palloc(p, fb->bufsiz + 2 + CHUNK_HEADER_SIZE + 1);
    else
	fb->outbase = NULL;

    fb->inptr = fb->inbase;

    fb->incnt = 0;
    fb->outcnt = 0;
    fb->outchunk = -1;
    fb->error = NULL;
    fb->bytes_sent = 0;
    fb->chunk_overcommit = 0;

    return fb;
}

static ap_status_t bcleanup(void *v)
{
    BUFF *fb = v;

    return iol_close(fb->iol);
}

/*
 * Push some I/O file descriptors onto the stream
 */
API_EXPORT(void) ap_bpush_iol(BUFF *fb, ap_iol *iol)
{
    fb->iol = iol;
    ap_register_cleanup(fb->pool, fb, bcleanup, bcleanup);
}

API_EXPORT(ap_status_t) ap_bsetopt(BUFF *fb, int optname, const void *optval)
{
    switch (optname) {
    case BO_BYTECT:
	fb->bytes_sent = *(const long int *) optval - (long int) fb->outcnt;
	return APR_SUCCESS;

    case BO_TIMEOUT:
	fb->flags &= ~B_NONBLOCK;
	if (*(const long int *)optval == 0) {
	    fb->flags |= B_NONBLOCK;
	    /* XXX: should remove B_WR now... */
	}
	return iol_setopt(fb->iol, AP_IOL_TIMEOUT, optval);
    }
    fb->berrno = APR_EINVAL;
    return APR_EINVAL;
}

API_EXPORT(ap_status_t) ap_bgetopt(BUFF *fb, int optname, void *optval)
{
    long int bs;

    switch (optname) {
    case BO_BYTECT:
	bs = fb->bytes_sent + fb->outcnt;
	if (bs < 0L)
	    bs = 0L;
	*(long int *) optval = bs;
	return APR_SUCCESS;

    case BO_TIMEOUT:
	return iol_getopt(fb->iol, AP_IOL_TIMEOUT, optval);
    }
    fb->berrno = APR_EINVAL;
    return APR_EINVAL;
}

/*
    start chunked encoding.

    in order for ap_bputc() to be an efficient macro we need to have
    the buffer setup for the chunk already.  if B_CHUNK is set, then
    routines here using end_chunk() must be sure to call start_chunk()
    or set an error condition before they return to the caller.
 */
static void start_chunk(BUFF *fb)
{
    fb->outchunk = fb->outcnt;
    fb->outcnt += CHUNK_HEADER_SIZE;
}

/*
    we've got a chunk in progress which we want to end -- the caller
    possibly wants to write extra bytes beyond those in fb->outbase.
    if extra != 0 then the caller is assumed to write the chunk CRLF
    trailer before calling start_chunk again.  there's more info below
    at large_write_chunk().
*/
static int end_chunk(BUFF *fb, int extra)
{
    int i;
    unsigned char *strp;
    int chunk_size;

    chunk_size = fb->outcnt - fb->outchunk - CHUNK_HEADER_SIZE + extra;
    if (chunk_size == 0) {
        /* nothing was written into this chunk, and we can't write a 0 size
         * chunk because that signifies EOF, so just erase it
         */
        fb->outcnt = fb->outchunk;
        fb->outchunk = -1;
        return 0;
    }

    if (chunk_size > MAX_CHUNK_SIZE) {
	extra -= chunk_size - MAX_CHUNK_SIZE;
	chunk_size = MAX_CHUNK_SIZE;
    }

    /* we know this will fit because of space we saved in start_chunk() */
    i = ap_snprintf((char *) &fb->outbase[fb->outchunk], CHUNK_HEADER_SIZE,
                "%x", chunk_size);

    /* we may have to tack some trailing spaces onto the number we just wrote
     * in case it was smaller than our estimated size.  We've also written
     * a \0 into the buffer with ap_snprintf so we might have to put a
     * \r back in.
     */
    strp = &fb->outbase[fb->outchunk + i];
    while (i < CHUNK_HEADER_SIZE - 2) {
        *strp++ = ' ';
        ++i;
    }
    *strp++ = '\015';
    *strp = '\012';

    if (extra == 0) {
	/* tack on the trailing CRLF, we've reserved room for this */
	fb->outbase[fb->outcnt++] = '\015';
	fb->outbase[fb->outcnt++] = '\012';
    }

    fb->outchunk = -1;
    
    return extra;
}



/*
 * Set a flag on (1) or off (0).
 */
API_EXPORT(int) ap_bsetflag(BUFF *fb, int flag, int value)
{
    int old_flags = fb->flags;

    if (value) {
	fb->flags |= flag;
	/* start chunking if we haven't already */
        if ((flag ^ old_flags) & B_CHUNK) {
            start_chunk(fb);
        }
    }
    else {
	fb->flags &= ~flag;
	/* stop chunking if we haven't already */
        if (flag & old_flags & B_CHUNK) {
            end_chunk(fb, 0);
        }
    }
    return value;
}


/* a wrapper around iol_read which checks for errors and EOFs */
static ap_status_t read_with_errors(BUFF *fb, void *buf, ap_size_t nbyte,
                                    ap_ssize_t *bytes_read)
{
    ap_status_t rv;

    rv = iol_read(fb->iol, buf, nbyte, bytes_read);
    if (rv == APR_SUCCESS && *bytes_read == 0) {
	fb->flags |= B_EOF;
    }
    else if (rv != APR_SUCCESS) {
	fb->berrno = rv;
	fb->saved_errno = rv;
	if (rv != APR_EAGAIN) {
	    doerror(fb, B_RD);
	}
    }
    return rv;
}


/*
 * Read up to nbyte bytes into buf.
 * If fewer than byte bytes are currently available, then return those.
 */
API_EXPORT(ap_status_t) ap_bread(BUFF *fb, void *buf, ap_size_t nbyte,
                                 ap_ssize_t *bytes_read)
{
    int nrd;
    ap_ssize_t i;
    ap_status_t rv;

    if (fb->flags & B_RDERR) {
        *bytes_read = 0;
	return fb->saved_errno;
    }
    if (nbyte == 0) {
        *bytes_read = 0;
	return APR_SUCCESS;
    }

    if (!(fb->flags & B_RD)) {
	/* Unbuffered reading.  First check if there was something in the
	 * buffer from before we went unbuffered. */
	if (fb->incnt) {
	    i = (fb->incnt > nbyte) ? nbyte : fb->incnt;
	    memcpy(buf, fb->inptr, i);
	    fb->incnt -= i;
	    fb->inptr += i;
            *bytes_read = i;
	    return APR_SUCCESS;
	}
	return read_with_errors(fb, buf, nbyte, bytes_read);
    }

    nrd = fb->incnt;
/* can we fill the buffer */
    if (nrd >= nbyte) {
	memcpy(buf, fb->inptr, nbyte);
	fb->incnt = nrd - nbyte;
	fb->inptr += nbyte;
        *bytes_read = nbyte;
	return APR_SUCCESS;
    }

    if (nrd > 0) {
	memcpy(buf, fb->inptr, nrd);
	nbyte -= nrd;
	buf = nrd + (char *) buf;
	fb->incnt = 0;
    }
    if (fb->flags & B_EOF) {
        *bytes_read = nrd;
	return APR_SUCCESS;
    }

/* do a single read */
    if (nbyte >= fb->bufsiz) {
/* read directly into caller's buffer */
	rv = read_with_errors(fb, buf, nbyte, &i);
	if (rv != APR_SUCCESS) {
            *bytes_read = nrd;
            return rv;
	}
    }
    else {
/* read into hold buffer, then memcpy */
	fb->inptr = fb->inbase;
	rv = read_with_errors(fb, fb->inptr, fb->bufsiz, &i);
	if (rv != APR_SUCCESS) {
            *bytes_read = nrd;
            return rv;
	}
	fb->incnt = i;
	if (i > nbyte)
	    i = nbyte;
	memcpy(buf, fb->inptr, i);
	fb->incnt -= i;
	fb->inptr += i;
    }
    *bytes_read = nrd + i;
    return APR_SUCCESS;
}


/*
 * Reads from the stream into the array pointed to by buff, until
 * a (CR)LF sequence is read, or end-of-file condition is encountered
 * or until n-1 bytes have been stored in buff. If a CRLF sequence is
 * read, it is replaced by a newline character.  The string is then
 * terminated with a null character.
 *
 * Returns the number of bytes stored in buff, or zero on end of
 * transmission, or -1 on an error.
 *
 * Notes:
 *  If null characters are expected in the data stream, then
 * buff should not be treated as a null terminated C string; instead
 * the returned count should be used to determine the length of the
 * string.
 *  CR characters in the byte stream not immediately followed by a LF
 * will be preserved.
 */
API_EXPORT(int) ap_bgets(char *buff, int n, BUFF *fb)
{
    int ch, ct;
    ap_ssize_t i;
    ap_status_t rv;

/* Can't do bgets on an unbuffered stream */
    if (!(fb->flags & B_RD)) {
	fb->berrno = APR_EINVAL;
	return -1;
    }
    if (fb->flags & B_RDERR) {
	return -1;
    }

    ct = 0;
    i = 0;
    for (;;) {
	if (i == fb->incnt) {
/* no characters left */
	    fb->inptr = fb->inbase;
	    fb->incnt = 0;
	    if (fb->flags & B_EOF)
		break;
	    rv = read_with_errors(fb, fb->inptr, fb->bufsiz, &i);
	    if (rv != APR_SUCCESS) {
		buff[ct] = '\0';
		return ct ? ct : -1;
	    }
	    fb->incnt = i;
	    if (i == 0)
		break;		/* EOF */
	    i = 0;
	    continue;		/* restart with the new data */
	}

	ch = fb->inptr[i++];
	if (ch == '\012') {	/* got LF */
	    if (ct == 0)
		buff[ct++] = '\n';
/* if just preceeded by CR, replace CR with LF */
	    else if (buff[ct - 1] == '\015')
		buff[ct - 1] = '\n';
	    else if (ct < n - 1)
		buff[ct++] = '\n';
	    else
		i--;		/* no room for LF */
	    break;
	}
	if (ct == n - 1) {
	    i--;		/* push back ch */
	    break;
	}

	buff[ct++] = ch;
    }
    fb->incnt -= i;
    fb->inptr += i;

    buff[ct] = '\0';
    return ct;
}

/*
 * Looks at the stream fb and places the first character into buff
 * without removing it from the stream buffer.
 *
 * Returns 1 on success, zero on end of transmission, or -1 on an error.
 *
 */
API_EXPORT(int) ap_blookc(BUFF *fb)
{
    int c;

    c = ap_bgetc(fb);
    if (c != -1) {
	ap_bungetc(c, fb);
    }
    return c;
}

/*
 * output a single character.  Used by ap_bputs when the buffer
 * is full... and so it'll cause the buffer to be flushed first.
 */
API_EXPORT(int) ap_bflsbuf(int c, BUFF *fb)
{
    char ss[1];
    ap_ssize_t bytes_written;
    ap_status_t rv;

    ss[0] = c;
    rv = ap_bwrite(fb, ss, 1, &bytes_written);
    if (rv == APR_SUCCESS) {
        return bytes_written;
    }
    return -1;
}

/*
 * Fill the buffer and read a character from it
 */
API_EXPORT(int) ap_bfilbuf(BUFF *fb)
{
    ap_status_t rv;
    ap_ssize_t i;
    char buf[1];

    rv = ap_bread(fb, buf, 1, &i);
    if (rv == APR_SUCCESS && i == 0)
        fb->berrno = APR_SUCCESS;       /* no error; EOF */
    if (i != 1)
	return EOF;
    else
	return buf[0];
}

/* A wrapper for writev which deals with error conditions and
 * bytes_sent.
 */
static ap_status_t writev_with_errors(BUFF *fb, const struct iovec *vec,
                                      int nvec, ap_ssize_t *bytes_written)
{
    ap_status_t rv;
    rv = iol_writev(fb->iol, vec, nvec, bytes_written);
    fb->bytes_sent += *bytes_written;
    if (rv != APR_SUCCESS) {
	fb->saved_errno = rv;
	if (rv != APR_EAGAIN) {
	    doerror(fb, B_WR);
	}
    }
    return rv;
}


static ap_status_t writev_it_all(BUFF *fb, struct iovec *vec, int nvec,
                                 ap_ssize_t *bytes_written)
{
    int i;
    ap_status_t rv;
    ap_ssize_t n;

    i = 0;
    *bytes_written = 0;
    while (i < nvec) {
	rv = writev_with_errors(fb, vec + i, nvec - i, &n);
	*bytes_written += n;
	if (rv != APR_SUCCESS) {
            return rv;
	}
	if (fb->flags & B_NONBLOCK) {
	    return APR_SUCCESS;
	}
	/* recalculate to deal with partial writes */
	while (n > 0) {
            if (n < vec[i].iov_len) {
                vec[i].iov_base = (char *) vec[i].iov_base + n;
                vec[i].iov_len -= n;
                break;
            }
            else {
                n -= vec[i].iov_len;
                ++i;
            }
        }
    }
    return APR_SUCCESS;
}


/* write the contents of fb->outbase, and buf,
   stop at first partial write for a non-blocking buff

   bytes_written is the number of bytes written from buf, it
   doesn't include any bytes potentially written from fb->outbase.
*/
static ap_status_t large_write(BUFF *fb, const char *buf, ap_size_t nbyte,
                               ap_ssize_t *bytes_written)
{
    struct iovec vec[2];
    int nvec;
    ap_status_t rv;

    ap_assert(nbyte > 0);
    if (fb->outcnt) {
	vec[0].iov_base = (void *)fb->outbase;
	vec[0].iov_len = fb->outcnt;
	vec[1].iov_base = (void *)buf;
	vec[1].iov_len = nbyte;
	nvec = 2;
    }
    else {
	vec[0].iov_base = (void *)buf;
	vec[0].iov_len = nbyte;
	nvec = 1;
    }
    rv = writev_it_all(fb, vec, nvec, bytes_written);
    if (*bytes_written < fb->outcnt) {
	/* shift bytes forward in buffer */
	fb->outcnt -= *bytes_written;
	memmove(fb->outbase, fb->outbase + *bytes_written, fb->outcnt);
        *bytes_written = 0;
	return rv;
    }
    *bytes_written -= fb->outcnt;
    fb->outcnt = 0;
    return rv;
}


/*
    large_write_chunk is similar to large_write(), but used when chunking
    is in effect.

    in apache-1.3 we used bcwrite() and large_write() when writing large
    chunks, and we did so in a way that caused us to write at least two
    small buffers into the iovec -- one for the chunk size, and one for
    the trailing CRLF.  (we had iovecs of size 3 or 4 when chunking)

    in 2.0 i wanted the ability to do non-blocking chunked i/o, which can
    result in partial writes at times.  this can be awfully inconvenient
    with the 1.3 method of chunking because there are a godawful number
    of special cases trying to remember how far into the iovec you got
    before EAGAIN.

    the insight i found was that we can combine the CRLF end of chunk
    N with the size of chunk N+1.  i do this right in the fb->outbase
    buffer, because any partial write of this chunk end/start data is
    the same as a partial write of any other buffered data.  this gets
    rid of almost all the special cases, and probably works out to be
    faster than letting the kernel handle a 2 byte and a 6 byte buffer.

    one special case is when we've told the client that we'd write a
    chunk of K bytes, and we get an EAGAIN before we write K bytes.
    i call this an "overcommit".  in this case we keep note of the fact
    that we've got an overcommit so that subsequent bwrites (which
    the layer above BUFF is performing in the non-blocking case) can
    decrement from the overcommit before they begin another chunk.

    -djg
*/
static ap_status_t large_write_chunk(BUFF *fb, const char *buf, ap_size_t nbyte,
                                     ap_ssize_t *bytes_written)
{
    ap_status_t rv;
    ap_ssize_t total, n, amt;

    ap_assert(nbyte > 0);
    total = 0;
    if (fb->chunk_overcommit) {
	amt = nbyte > fb->chunk_overcommit ? fb->chunk_overcommit : nbyte;
	rv = large_write(fb, buf, amt, &total);
	fb->chunk_overcommit -= total;
	if (rv != APR_SUCCESS) {
	    *bytes_written = total;
            return rv;
	}
	if (fb->chunk_overcommit) {
	    /* above we ensured that we'd write at most the overcommit
	       bytes -- so either we had a partial write,
	       or we used up all of buf.  either way we're done.  */
	    ap_assert(amt < fb->chunk_overcommit || *bytes_written < amt);
	    *bytes_written = total;
	    return APR_SUCCESS;
	}
	/* we've finished the overcommit so we end off the previous
	   chunk, begin a new one, and shift buf */
	fb->outbase[0] = '\015';
	fb->outbase[1] = '\012';
	fb->outcnt = 2;
	start_chunk(fb);
	nbyte -= total;
	buf += total;
    }
    ap_assert(fb->chunk_overcommit == 0 && fb->outchunk != -1);
    while (nbyte) {
	amt = end_chunk(fb, nbyte);
	rv = large_write(fb, buf, amt, &n);
	if (n < amt) {
	    /* can only be non-blocking or an error... */
	    fb->chunk_overcommit = amt - n;
            *bytes_written = total + n;
            return rv;
	}
	fb->outbase[0] = '\015';
	fb->outbase[1] = '\012';
	fb->outcnt = 2;
	start_chunk(fb);
	nbyte -= amt;
	buf += amt;
	total += amt;
    }
    *bytes_written = total;
    return APR_SUCCESS;
}

/* A wrapper for write which deals with error conditions and
 * bytes_sent.
 */
static ap_status_t write_with_errors(BUFF *fb, const void *buf, ap_size_t nbyte,
                                     ap_ssize_t *bytes_written)
{
    ap_status_t rv;

    rv = iol_write(fb->iol, buf, nbyte, bytes_written);
    fb->bytes_sent += *bytes_written;
    if (rv != APR_SUCCESS) {
	fb->saved_errno = rv;
	if (rv != APR_EAGAIN) {
	    doerror(fb, B_WR);
	}
    }
    return rv;
}


static ap_status_t bflush_core(BUFF *fb, ap_ssize_t *bytes_written)
{
    ap_status_t rv;
    ap_ssize_t n;

    if (fb->flags & B_CHUNK) {
	end_chunk(fb, 0);
    }
    *bytes_written = 0;
    while (fb->outcnt > 0) {
	rv = write_with_errors(fb, fb->outbase + *bytes_written, fb->outcnt,
                                 &n);
	if (rv != APR_SUCCESS) {
	    if (*bytes_written) {
		memmove(fb->outbase, fb->outbase + *bytes_written, fb->outcnt);
	    }
	    return rv;
	}
	fb->outcnt -= n;
	*bytes_written += n;
    }
    if (fb->flags & B_CHUNK) {
	start_chunk(fb);
    }
    return APR_SUCCESS;
}


/*
 * Write nbyte bytes.
 * Only returns fewer than nbyte if an error ocurred.
 * Returns -1 if no bytes were written before the error ocurred.
 * It is worth noting that if an error occurs, the buffer is in an unknown
 * state.
 */
API_EXPORT(ap_status_t) ap_bwrite(BUFF *fb, const void *buf, ap_size_t nbyte,
                                  ap_ssize_t *bytes_written)
{
    int amt;
    int total;
    ap_ssize_t n;
    ap_status_t rv;

    if (fb->flags & (B_WRERR | B_EOUT)) {
        *bytes_written = 0;
	return fb->saved_errno;
    }
    if (nbyte == 0) {
        *bytes_written = 0;
	return APR_SUCCESS;
    }

/*
 * Detect case where we're asked to write a large buffer, and combine our
 * current buffer with it in a single writev().  Note we don't consider
 * the case nbyte == 1 because modules which use rputc() loops will cause
 * us to use writev() too frequently.  In those cases we really should just
 * start a new buffer.
 */
    if (!(fb->flags & B_WR)
	|| (nbyte > LARGE_WRITE_THRESHOLD && nbyte + fb->outcnt >= fb->bufsiz)) {
	if (fb->flags & B_CHUNK) {
	    return large_write_chunk(fb, buf, nbyte, bytes_written);
	}
        else {
	    return large_write(fb, buf, nbyte, bytes_written);
        }
    }

    /* at this point we know that nbyte < fb->bufsize */
    amt = fb->bufsiz - fb->outcnt;
    total = 0;
    if (nbyte > amt) {
	memcpy(fb->outbase + fb->outcnt, buf, amt);
	fb->outcnt += amt;
	buf = (const char *) buf + amt;
	nbyte -= amt;
        rv = bflush_core(fb, &n);
	if (rv != APR_SUCCESS) {
	    /* don't continue past an error or a partial write */
	    /* regardless of how much flushed we've at least copied amt
	       into our buffers... */
            *bytes_written = amt;
            return APR_SUCCESS;
	}
	total = amt;
    }
    /* now we know that nbyte < fb->bufsiz */
    memcpy(fb->outbase + fb->outcnt, buf, nbyte);
    fb->outcnt += nbyte;
    *bytes_written = total + nbyte;
    return APR_SUCCESS;
}

/*
 * Flushes the buffered stream.
 */
API_EXPORT(ap_status_t) ap_bflush(BUFF *fb)
{
    ap_ssize_t n;       /* Placeholder; not ever used */

    if ((fb->flags & (B_EOUT | B_WR)) != B_WR) {
        fb->berrno = APR_EINVAL;
        return APR_EINVAL;
    }
    if ((fb->flags & B_WRERR) != 0) {
        return fb->saved_errno;
    }
    return bflush_core(fb, &n);
}

/*
 * Flushes and closes the file, even if an error occurred.
 * Discards an data that was not read, or not written by bflush()
 * Sets the EOF flag to indicate no futher data can be read,
 * and the EOUT flag to indicate no further data can be written.
 */
API_EXPORT(ap_status_t) ap_bclose(BUFF *fb)
{
    ap_status_t rc1, rc2;

    if (fb->flags & B_WR)
	rc1 = ap_bflush(fb);
    else
	rc1 = APR_SUCCESS;
    ap_kill_cleanup(fb->pool, fb, bcleanup);
    rc2 = iol_close(fb->iol);

    fb->inptr = fb->inbase;
    fb->incnt = 0;
    fb->outcnt = 0;

    fb->flags |= B_EOF | B_EOUT;

    if (rc1 != APR_SUCCESS)
	return rc1;
    return rc2;
}

/*
 * returns the number of bytes written or -1 on error
 */
API_EXPORT(int) ap_bputs(const char *x, BUFF *fb)
{
    ap_status_t rv;
    int j = strlen(x);
    ap_ssize_t i;
    rv = ap_bwrite(fb, x, j, &i);
    if (i != j) {
        fb->berrno = rv;
	return -1;
    }
    else
	return j;
}

/*
 * returns the number of bytes written or -1 on error
 */
API_EXPORT_NONSTD(int) ap_bvputs(BUFF *fb,...)
{
    int j, k;
    ap_ssize_t i;
    va_list v;
    const char *x;

    va_start(v, fb);
    for (k = 0;;) {
	x = va_arg(v, const char *);
	if (x == NULL)
	    break;
	j = strlen(x);
	(void) ap_bwrite(fb, x, j, &i);
	if (i != j) {
	    va_end(v);
	    return -1;
	}
	k += i;
    }

    va_end(v);

    return k;
}

API_EXPORT(void) ap_bonerror(BUFF *fb,
                             void (*error) (BUFF *, int, void *, ap_status_t),
                             void *data)
{
    fb->error = error;
    fb->error_data = data;
}

struct bprintf_data {
    ap_vformatter_buff_t vbuff;
    BUFF *fb;
};

API_EXPORT(ap_status_t) ap_berror(BUFF *fb)
{
    return fb->berrno;
}

static int bprintf_flush(ap_vformatter_buff_t *vbuff)
{
    struct bprintf_data *b = (struct bprintf_data *)vbuff;
    BUFF *fb = b->fb;

    fb->outcnt += b->vbuff.curpos - (char *)&fb->outbase[fb->outcnt];
    if (fb->outcnt == fb->bufsiz) {
	if (ap_bflush(fb) != APR_SUCCESS) {
	    return -1;
	}
    }
    vbuff->curpos = (char *)&fb->outbase[fb->outcnt];
    vbuff->endpos = (char *)&fb->outbase[fb->bufsiz];
    return 0;
}

API_EXPORT_NONSTD(int) ap_bprintf(BUFF *fb, const char *fmt, ...)
{
    va_list ap;
    int res;
    struct bprintf_data b;

    /* XXX: only works with buffered writes */
    if ((fb->flags & (B_WRERR | B_EOUT | B_WR)) != B_WR) {
        fb->berrno = APR_EINVAL;
	return -1;
    }
    b.vbuff.curpos = (char *)&fb->outbase[fb->outcnt];
    b.vbuff.endpos = (char *)&fb->outbase[fb->bufsiz];
    b.fb = fb;
    va_start(ap, fmt);
    res = ap_vformatter(bprintf_flush, &b.vbuff, fmt, ap);
    va_end(ap);
    if (res != -1) {
	fb->outcnt += b.vbuff.curpos - (char *)&fb->outbase[fb->outcnt];
    }
    return res;
}

API_EXPORT(int) ap_vbprintf(BUFF *fb, const char *fmt, va_list ap)
{
    struct bprintf_data b;
    int res;

    /* XXX: only works with buffered writes */
    if ((fb->flags & (B_WRERR | B_EOUT | B_WR)) != B_WR) {
        fb->berrno = APR_EINVAL;
	return -1;
    }
    b.vbuff.curpos = (char *)&fb->outbase[fb->outcnt];
    b.vbuff.endpos = (char *)&fb->outbase[fb->bufsiz];
    b.fb = fb;
    res = ap_vformatter(bprintf_flush, &b.vbuff, fmt, ap);
    if (res != -1) {
	fb->outcnt += b.vbuff.curpos - (char *)&fb->outbase[fb->outcnt];
    }
    return res;
}

API_EXPORT(BUFF *) ap_bopenf(ap_context_t *a, const char *name, int flg, int mode)
{
    ap_iol *iol;
    BUFF *fb;
    ap_status_t rv;
    ap_file_t *file = NULL;
    rv = ap_open(&file, name, flg, 0, a);
    if ((rv != APR_SUCCESS) || (file == NULL)) {
        return NULL;
    }
    iol = ap_create_file_iol(file);
    if (!iol) {
        return NULL;
    }
    fb = ap_bcreate(a, ((flg &(B_RD|B_RDWR)) ? B_RD : 0)
		     | ((flg & (B_WR|B_RDWR)) ? B_WR : 0));
    ap_bpush_iol(fb, iol);
    return fb;
}
